// MCE_VR_GLUT by Michal Husak husakm@vscht.cz
// version 11 final 

#include <math.h>
#include <stdlib.h>

// swich on/off CAVE emulation mode
//#define CAVE_EMULATION

#ifdef CAVE_EMULATION
	#include "cave_ogl_emul.h"
#else
	#include "cave_ogl.h"
        #include "glut.h"
        #include <unistd.h>
#endif

//#include <GL/gl.h>	// Header File For The OpenGL32 Library
//#include <GL/glu.h>	// Header File For The GLu32 Library
#include <stdio.h>      // Header file for standard file i/o.
//#include <stdlib.h>     // Header file for malloc/free.
//#include <math.h>       // Header file for trigonometric functions.
#include <string.h> // string operations

// Michal Husak VFX3D
#include "resource.h"
#include "MCE_VR_CAVE.h"
#include "log.h"
#include "mcube_bourke.h"
#include "voxel_procesing.h"
#include "molecular_objects.h"
#include "atom_procesing.h"
#include <float.h> // float number limits

// Michal Husak fps calulation
#include <time.h>

// ascii codes for various special keys
#define ESCAPE 27
#define PAGE_UP 73
#define PAGE_DOWN 81
#define UP_ARROW 72
#define DOWN_ARROW 80
#define LEFT_ARROW 75
#define RIGHT_ARROW 77

const float piover180 = 0.0174532925f;

//GLfloat z=0.0f;                       // depth into the screen.

ELD_MapData MapData;
MoleculeData MolData;


// CAVE Shared data
float *x_cave,*y_cave,*z_cave;
float *x_cross,*y_cross,*z_cross;
float *al_cave,*be_cave,*ga_cave;
float *model_scale;
float *zero;
int *need_dl_recalc;
CaveMode *CurrentCaveMode;
int *layer_visibility;

MceVrMode	*Mode;
MceVrConfiguration *Conf;


MoleculeData *PeakData;

// data witch must persist the CAVE deinitialization (IRIX specific problem)

typedef struct {
float PlaneMatrix[3][3]; //float PlaneMatrix[3][3];
float PlaneTranslation[3]; //float PlaneTranslation[3];
} MapDataShared;

MapDataShared *MapDataS;

// displaylists symbolic names
// RESERVED = 0 , non correct name !!!
typedef enum {RESERVED,MCE_TRIANGLES,MCE_CONTOURS,MCE_CONTOURS2D,MCE_MOLECULE} displayLists;

// triangle field
static unsigned int g_iNtriangles[3];
static TRIANGLE triangles[3][MAXTRIANGLES];

// line controur field
static unsigned int g_iNcontourlines[3];
static LINE	contourlines[3][MAXCONTOURLINES];

static float layer_limits[3];
#define START 0.1
#define STEP 0.2

// degrees to radians...2 PI radians = 360 degrees
float rad(float angle)
{
    return angle * piover180;
}

// helper for SetupWorld.  reads a file into a string until a nonblank, non-comment line
// is found ("/" at the start indicating a comment); assumes lines < 255 characters long.
void readstr(FILE *f, char *string)
{
    do {
	fgets(string, 255, f); // read the line
    } while ((string[0] == '/') || (string[0] == '\n'));
    return;
}

void ReadFouFile()
{
	FILE *filein;
	char oneline[255];
	// debuging
#ifdef _DEBUG
	char szBuffer[256];
#endif

	// support variable
	int i, j;
	int x,y,z;
	int atom_record_lenght;

	filein = fopen("Data/MCE_DATA.fou", "rt");				// File To Load World Data From

 	if (filein == NULL) LogMsg("No Data/Test.fou file fiound\n");
	readstr(filein,oneline); //INFO DOWN, ....?
	readstr(filein,oneline); //TRAN ?

	for (i=0;i<3;i++)
	{
		for(j=0;j<3;j++)
		{
			readstr(filein,oneline);
			MapData.PlaneMatrix[i][j]=(float) atof(oneline);
			MapDataS->PlaneMatrix[i][j]=(float) atof(oneline);
#ifdef _DEBUG
			sprintf(szBuffer,"%d %d %f \n",i,j,MapDataS->PlaneMatrix[i][j]);

			LogMsg(szBuffer);
#endif
		}
	}
	
	for (i=0;i<3;i++)
	{
		readstr(filein,oneline);
		MapData.PlaneTranslation[i]=(float) atof(oneline);
		MapDataS->PlaneTranslation[i]=(float) atof(oneline);
	}

	readstr(filein,oneline); // CELL

	for (i=0;i<6;i++)
	{
		readstr(filein,oneline);
		MapData.Cell[i]=(float) atof(oneline);
	}

	readstr(filein,oneline); // L14

	for (i=0;i<3;i++)
	{
		readstr(filein,oneline);
		MapData.AxisParam[i].Start=(float) atof(oneline);

		readstr(filein,oneline);
		MapData.AxisParam[i].Step=(float) atof(oneline);

		readstr(filein,oneline);
		MapData.AxisParam[i].End=(float) atof(oneline);

		readstr(filein,oneline);
		MapData.AxisParam[i].Division=(float) atof(oneline);
	}

	readstr(filein,oneline); // SIZE

	readstr(filein,oneline);
	MapData.Lines=atoi(oneline);

	readstr(filein,oneline);
	MapData.Columns=atoi(oneline);

	readstr(filein,oneline);
	MapData.Sections=atoi(oneline);

	// !!! This parto of the code is not memory-optimal !!!
	// it should be something like in this place 
	// ReDim Voxel(MapData.Lines, MapData.Columns, MapData.Sections)

	for (z=0;z<MapData.Sections;z++)
	{
		readstr(filein,oneline); // BLOCK
		readstr(filein,oneline); // N

#ifdef _DEBUG
			sprintf(szBuffer,"Z layer: %d Values: %s\n",z,oneline);
			LogMsg(szBuffer);
#endif

		for(x=0;x<MapData.Lines;x++)
		{
			for(y=0;y<MapData.Columns;y++)
			{
				readstr(filein,oneline);
				MapData.Voxel[x][y][z]=(float) atof(oneline);
			}
		}
	}

	// molecular data input

	readstr(filein,oneline); //LIST5

	readstr(filein,oneline);
	MolData.Natoms=atoi(oneline);

	if (MolData.Natoms>0)
	{
		readstr(filein,oneline);
		atom_record_lenght=atoi(oneline);
		for (i=0;i<MolData.Natoms;i++)
		{
			readstr(filein,oneline); // atom name
			strncpy(MolData.a[i].name,oneline,2);

			readstr(filein,oneline);
			MolData.a[i].atm_label_number=atoi(oneline);

			readstr(filein,oneline);
			MolData.a[i].occ=(float) atof(oneline);

			readstr(filein,oneline); // reserved, dummy now

			readstr(filein,oneline);
			MolData.a[i].fc.x=(float) atof(oneline);

			readstr(filein,oneline);
			MolData.a[i].fc.y=(float) atof(oneline);

			readstr(filein,oneline);
			MolData.a[i].fc.z=(float) atof(oneline);
			
			readstr(filein,oneline);
			MolData.a[i].uiso=(float) atof(oneline);
			
			for(j=0;j<6;j++)
			{
				readstr(filein,oneline);
				MolData.a[i].u[j]=(float) atof(oneline);
			}

			// dumy data from future fou versions
			if (atom_record_lenght>14)
			{
				for(j=15;j<=atom_record_lenght;j++)
				{
					readstr(filein,oneline); // dummy reading
				}
			}
		}// next j
	} // endif atom presents

	fclose(filein);
}

XYZ triangle_center(XYZ p0,XYZ p1, XYZ p2)
{
	XYZ result ;

	result.x=(p0.x+p1.x+p2.x)/3;
	result.y=(p0.y+p1.y+p2.y)/3;
	result.z=(p0.z+p1.z+p2.z)/3;

	return result;
}

void GenerateDisplayLists()
{
	unsigned long i;
	XYZ center;
	// MCE line object generation
#ifdef _DEBUG
	LogMsg("Display List generation started\n");
#endif
	
	switch(Mode->GridType)
	{
	case MCube_Triangle:

	glNewList(MCE_CONTOURS, GL_COMPILE);

	if(layer_visibility[0])
	{
		glColor3f (0.0f, 1.0f, 0.0f); // green lines
		for (i=0;i<g_iNtriangles[0];i++)
		{
			glBegin(GL_LINE_LOOP);
				glVertex3f(triangles[0][i].p[0].x,triangles[0][i].p[0].y,triangles[0][i].p[0].z);
				glVertex3f(triangles[0][i].p[1].x,triangles[0][i].p[1].y,triangles[0][i].p[1].z);
				glVertex3f(triangles[0][i].p[2].x,triangles[0][i].p[2].y,triangles[0][i].p[2].z);
			glEnd();
		}
	}
	if(layer_visibility[1])
	{
		glColor3f (0.0f, 0.0f, 1.0f); // blue lines
		for (i=0;i<g_iNtriangles[1];i++)
		{
			glBegin(GL_LINE_LOOP);
				glVertex3f(triangles[1][i].p[0].x,triangles[1][i].p[0].y,triangles[1][i].p[0].z);
				glVertex3f(triangles[1][i].p[1].x,triangles[1][i].p[1].y,triangles[1][i].p[1].z);
				glVertex3f(triangles[1][i].p[2].x,triangles[1][i].p[2].y,triangles[1][i].p[2].z);
			glEnd();
		}
	}
	if(layer_visibility[2])
	{
		glColor3f (1.0f, 0.0f, 0.0f); // red lines
		for (i=0;i<g_iNtriangles[2];i++)
		{
			glBegin(GL_LINE_LOOP);
				glVertex3f(triangles[2][i].p[0].x,triangles[2][i].p[0].y,triangles[2][i].p[0].z);
				glVertex3f(triangles[2][i].p[1].x,triangles[2][i].p[1].y,triangles[2][i].p[1].z);
				glVertex3f(triangles[2][i].p[2].x,triangles[2][i].p[2].y,triangles[2][i].p[2].z);
			glEnd();
		}
	}
	glEndList();
	break;

	case MCube_Web:
	glNewList(MCE_CONTOURS, GL_COMPILE);

		glBegin(GL_LINES);
		if(layer_visibility[0])
		{
		glColor3f (0.0f, 1.0f, 0.0f); // green lines
		for (i=0;i<g_iNtriangles[0];i++)
		{
			center=triangle_center(triangles[0][i].p[0],triangles[0][i].p[1],triangles[0][i].p[2]);

			glVertex3f(triangles[0][i].p[0].x,triangles[0][i].p[0].y,triangles[0][i].p[0].z);
			glVertex3f(center.x,center.y,center.z);
			glVertex3f(triangles[0][i].p[1].x,triangles[0][i].p[1].y,triangles[0][i].p[1].z);
			glVertex3f(center.x,center.y,center.z);
			glVertex3f(triangles[0][i].p[2].x,triangles[0][i].p[2].y,triangles[0][i].p[2].z);
			glVertex3f(center.x,center.y,center.z);
		}
		}
		if(layer_visibility[1])
		{
		glColor3f (0.0f, 0.0f, 1.0f); // blue lines
		for (i=0;i<g_iNtriangles[1];i++)
		{
			center=triangle_center(triangles[1][i].p[0],triangles[1][i].p[1],triangles[1][i].p[2]);

			glVertex3f(triangles[1][i].p[0].x,triangles[1][i].p[0].y,triangles[1][i].p[0].z);
			glVertex3f(center.x,center.y,center.z);
			glVertex3f(triangles[1][i].p[1].x,triangles[1][i].p[1].y,triangles[1][i].p[1].z);
			glVertex3f(center.x,center.y,center.z);
			glVertex3f(triangles[1][i].p[2].x,triangles[1][i].p[2].y,triangles[1][i].p[2].z);
			glVertex3f(center.x,center.y,center.z);
		}
		}
		if(layer_visibility[2])
		{
		glColor3f (1.0f, 0.0f, 0.0f); // red lines
		for (i=0;i<g_iNtriangles[2];i++)
		{
			center=triangle_center(triangles[2][i].p[0],triangles[2][i].p[1],triangles[2][i].p[2]);

			glVertex3f(triangles[2][i].p[0].x,triangles[2][i].p[0].y,triangles[2][i].p[0].z);
			glVertex3f(center.x,center.y,center.z);
			glVertex3f(triangles[2][i].p[1].x,triangles[2][i].p[1].y,triangles[2][i].p[1].z);
			glVertex3f(center.x,center.y,center.z);
			glVertex3f(triangles[2][i].p[2].x,triangles[2][i].p[2].y,triangles[2][i].p[2].z);
			glVertex3f(center.x,center.y,center.z);
		}
		}
		glEnd();
	glEndList();
	break;

	case MSquare_XYZ:
	case MSquare_Z:
	case MSquare_XZ:
	// line contour generation
	glNewList(MCE_CONTOURS, GL_COMPILE);

		glBegin(GL_LINES);
		if(layer_visibility[0])
		{
		glColor3f (0.0f, 1.0f, 0.0f); // green lines
		for (i=0;i<g_iNcontourlines[0];i++)
		{			
				glVertex3f(contourlines[0][i].p[0].x,contourlines[0][i].p[0].y,contourlines[0][i].p[0].z);
				glVertex3f(contourlines[0][i].p[1].x,contourlines[0][i].p[1].y,contourlines[0][i].p[1].z);
		}
		}
		if(layer_visibility[1])
		{
		glColor3f (0.0f, 0.0f, 1.0f); // blue lines
		for (i=0;i<g_iNcontourlines[1];i++)
		{
				glVertex3f(contourlines[1][i].p[0].x,contourlines[1][i].p[0].y,contourlines[1][i].p[0].z);
				glVertex3f(contourlines[1][i].p[1].x,contourlines[1][i].p[1].y,contourlines[1][i].p[1].z);
		}
		}
		if(layer_visibility[2])
		{
		glColor3f (1.0f, 0.0f, 0.0f); // red lines
		for (i=0;i<g_iNcontourlines[2];i++)
		{
				glVertex3f(contourlines[2][i].p[0].x,contourlines[2][i].p[0].y,contourlines[2][i].p[0].z);
				glVertex3f(contourlines[2][i].p[1].x,contourlines[2][i].p[1].y,contourlines[2][i].p[1].z);
		}
		}
		glEnd();
	
	glEndList();
	break;
	} // end switch
#ifdef _DEBUG
	LogMsg("Display List generation finished \n");
#endif
}


float distance (XYZ p1, XYZ p2)
{
// simple distance of two points

float d1,d2,d3;

	d1=p1.x-p2.x;
	d2=p1.y-p2.y;
	d3=p1.z-p2.z;

return sqrt(d1*d1+d2*d2+d3*d3);

}

void SetupConectivity()
{
 int i, j;
 float blenght;

 MolData.bond.nBond=0;

 for(i=0; i<MolData.Natoms; i++)
 {
	 for(j=i+1; j<MolData.Natoms;j++)
	 {
		 blenght=distance(MolData.a[i].cc,MolData.a[j].cc);

		 // shuld be replaced by property table int the future
		 if((strcmp(MolData.a[i].name,"H ")==0) | (strcmp(MolData.a[j].name,"H ")==0))
		 {
			if((blenght>0.7)&&(blenght<1.2)) // new atom
			{
				MolData.bond.nIndex[MolData.bond.nBond][0]=i;
				MolData.bond.nIndex[MolData.bond.nBond][1]=j;
				MolData.bond.nBond=MolData.bond.nBond+1;
			}
		 }
		 else if((blenght>0.7)&&(blenght<1.65)) // new atom
		 {		
			MolData.bond.nIndex[MolData.bond.nBond][0]=i;
			MolData.bond.nIndex[MolData.bond.nBond][1]=j;
			MolData.bond.nBond=MolData.bond.nBond+1;
		 }
	
	 } // next j
 }

}

void GenerateMesh()
{
	int i;

#ifdef _DEBUG
	LogMsg("Mesh generation\n");
#endif

	// initiliaze the number of graphic element values
	for(i=0;i<=2;i++)
	{
		g_iNcontourlines[i]=0;
		g_iNtriangles[i]=0;
	}
	
	switch(Mode->GridType)
	{
		case MSquare_XYZ:
		{
			VoxelDirectionalContour(&g_iNcontourlines[0],&MapData,layer_limits,contourlines,X);
			VoxelDirectionalContour(&g_iNcontourlines[0],&MapData,layer_limits,contourlines,Y);
			VoxelDirectionalContour(&g_iNcontourlines[0],&MapData,layer_limits,contourlines,Z);
		}
		break;

		case MSquare_Z:
		{
			VoxelDirectionalContour(&g_iNcontourlines[0],&MapData,layer_limits,contourlines,Z);
		}
		break;

		case MSquare_XZ:
		{
			VoxelDirectionalContour(&g_iNcontourlines[0],&MapData,layer_limits,contourlines,X);
			VoxelDirectionalContour(&g_iNcontourlines[0],&MapData,layer_limits,contourlines,Z);
		}
		break;

		case MCube_Triangle:
		{
			GenerateTrianglesFromVoxels(&g_iNtriangles[0],&MapData,layer_limits,triangles);
		}
		break;

		case MCube_Web:
		{
			GenerateTrianglesFromVoxels(&g_iNtriangles[0],&MapData,layer_limits,triangles);
		}
		break;	
	}
#ifdef _DEBUG
	LogMsg("Mesh generation completly finished\n");
#endif
}

void SetupWorld()
{
	MolData.Natoms=0;

	ReadFouFile();

	StuctureToCart(&MolData,&MapData);

	SetupConectivity();

	SetLayerLimits((float)START,(float)STEP,&layer_limits[0],&MapData);

	GenerateMesh();
}

void SetAtomColor(char name[2])
{
	if (strncmp(name,"C ",2)==0)
	{
		glColor3f(0.7f,0.7f,0.7f); // gray
	}

	else if (strncmp(name,"H ",2)==0)
	{
		glColor3f(0.7f,0.7f,1.0f); // light blue
	}

	else if (strncmp(name,"N ",2)==0)
	{
		glColor3f(0.0f,0.0f,0.8f); // blue
	}

	else if (strncmp(name,"O ",2)==0)
	{
		glColor3f(0.8f,0.0f,0.0f); // red
	}

	else if (strncmp(name,"S ",2)==0)
	{
		glColor3f(0.8f,0.8f,0.0f); // yellow
	}

	else
	{
		glColor3f(0.4f,0.4f,0.4f); // dark gray
	}
}


void DrawAtoms()
{
	int i;

	static GLUquadricObj *quadObj3=NULL;

	// sphere matereial
	GLfloat mat_specular[]={1.0,1.0,1.0,1.0};
	GLfloat mat_shiness[]={50.0};

	glMaterialfv(GL_FRONT,GL_SPECULAR,mat_specular);
	glMaterialfv(GL_FRONT,GL_SHININESS,mat_shiness);
	
  // sphere 
	if (! quadObj3)
	{
		quadObj3 = gluNewQuadric();
		//object properties (invisible if not defined !)
		//type of filling LINE,FILL
		gluQuadricDrawStyle (quadObj3, GLU_FILL);
		//type of shading - SMOOTH,FLAT
		gluQuadricNormals (quadObj3, GLU_SMOOTH);
		//vektors INSIDE OUTSIDE - for light calculations ??
		gluQuadricOrientation (quadObj3, GLU_OUTSIDE);
	}
	
	for(i=0;i<MolData.Natoms;i++)
	{
	glPushMatrix();

		SetAtomColor(MolData.a[i].name);

		glTranslatef(MolData.a[i].cc.x,
					 MolData.a[i].cc.y,
				     MolData.a[i].cc.z);

		// shuld be replaced by radius table in the future ...
		
		if (strcmp(MolData.a[i].name,"H ")==0)
		{
			gluSphere (quadObj3, 0.1 ,8,8);
		}
		else
		{
			gluSphere (quadObj3, 0.2 ,8,8);
		}

	glPopMatrix();
	}
}

void DrawBonds()
{
	int i;
	int i1;
	int i2;

	glColor3f(1,1,1); // white lines for bonds

	

	for(i=0;i<MolData.bond.nBond;i++)
	{
		i1=MolData.bond.nIndex[i][0];
		i2=MolData.bond.nIndex[i][1];

	glBegin(GL_LINES);
		glVertex3f(MolData.a[i1].cc.x,MolData.a[i1].cc.y,MolData.a[i1].cc.z);
		glVertex3f(MolData.a[i2].cc.x,MolData.a[i2].cc.y,MolData.a[i2].cc.z);
	glEnd();

	}	
}

void DrawPeaks()
{
	int i;

	static GLUquadricObj *quadObj2=NULL;

	// sphere matereial
	GLfloat mat_specular[]={1.0,1.0,1.0,1.0};
	GLfloat mat_shiness[]={50.0};

	glMaterialfv(GL_FRONT,GL_SPECULAR,mat_specular);
	glMaterialfv(GL_FRONT,GL_SHININESS,mat_shiness);
	
  // sphere 
	if (! quadObj2)
	{
		quadObj2 = gluNewQuadric();
		//object properties (invisible if not defined !)
		//type of filling LINE,FILL
		gluQuadricDrawStyle (quadObj2, GLU_FILL);
		//type of shading - SMOOTH,FLAT
		gluQuadricNormals (quadObj2, GLU_SMOOTH);
		//vektors INSIDE OUTSIDE - for light calculations ??
		gluQuadricOrientation (quadObj2, GLU_OUTSIDE);
	}
	
	for(i=0;i<PeakData->Natoms;i++)
	{
	glPushMatrix();

		glColor3f(1,1,0) ; //yellow  ...

		glTranslatef(PeakData->a[i].cc.x,
					 PeakData->a[i].cc.y,
				     PeakData->a[i].cc.z);

		gluSphere (quadObj2, 0.20 ,8,8);

	glPopMatrix();
	}
}

void NavigationCross()
{
float size=0.5;

static GLfloat rot_angle=0;

rot_angle+=1;

if (rot_angle > 360) rot_angle =0;

	glPushMatrix();
	glRotatef(rot_angle,1,1,1);
	glBegin(GL_LINES);
		glColor4f(1,1,1,1);
		glVertex3f(-size,0,0);
		glVertex3f(size,0,0);
		glVertex3f(0,-size,0);
		glVertex3f(0,size,0);
		glVertex3f(0,0,+size);
		glVertex3f(0,0,-size);
	glEnd();

	glPopMatrix();
}

GLvoid DrawCAVEScene(GLvoid)	// Here's Where We Do All The Drawing
{		
	glLineWidth(Mode->LineWidth); // nesesery for ticknes setup				

	glEnable(GL_LIGHTING);

		DrawAtoms();

		DrawPeaks();

	glDisable(GL_LIGHTING);

	DrawBonds();

	glCallList(MCE_CONTOURS);
}

void SetDefaultConf()
{
	FILE *filein;
	char oneline[255];
	char code[255];
	
	filein = fopen("Data/MCE_VR.CFG", "rt");
	
	// default setting

	Conf->ControlMode=IDC_CONTROL_KEYBOARD;
	Conf->LineWidthCode=IDC_LINE1;
	Conf->OSD_mode=IDC_OSD_YES;
	Conf->GridSelection=IDC_M_SQUARE_XYZ;

	if (filein==NULL)
	{
		LogMsg("MCE_VR.CFG not found, using default configuration.");			
	}
	else
	{			
		// interpreting MCE_VR.CFG file
		
		readstr(filein,oneline);
		sscanf(oneline, "ControlMode %s\n", code);
		if (strstr(code,"IDC_CONTROL_KEYBOARD")) Conf->ControlMode=IDC_CONTROL_KEYBOARD;
		if (strstr(code,"IDC_CONTROL_JOYSTICK")) Conf->ControlMode=IDC_CONTROL_JOYSTICK;
		if (strstr(code,"IDC_CONTROL_VFX3D_KEYBOARD")) Conf->ControlMode=IDC_CONTROL_VFX3D_KEYBOARD;
		if (strstr(code,"IDC_CONTROL_VFX3D_CYBERPUCK")) Conf->ControlMode=IDC_CONTROL_VFX3D_CYBERPUCK;
		if (strstr(code,"IDC_CONTROL_VFX3D_JOYPAD")) Conf->ControlMode=IDC_CONTROL_VFX3D_JOYPAD;
			
		readstr(filein,oneline);
		sscanf(oneline, "LineWidthCode %s\n", code);
		if (strstr(code,"IDC_LINE1")) 	Conf->LineWidthCode=IDC_LINE1;
		if (strstr(code,"IDC_LINE2")) 	Conf->LineWidthCode=IDC_LINE2;
		if (strstr(code,"IDC_LINE3")) 	Conf->LineWidthCode=IDC_LINE3;
	
		readstr(filein,oneline);
		sscanf(oneline, "OSD_mode %s\n", code);
		if (strstr(code,"IDC_OSD_YES")) Conf->OSD_mode=IDC_OSD_YES;
		if (strstr(code,"IDC_OSD_NO")) 	Conf->OSD_mode=IDC_OSD_NO;
	
		readstr(filein,oneline);
		sscanf(oneline, "GridSelection %s\n", code);
		if (strstr(code,"IDC_M_SQUARE_XYZ")) Conf->GridSelection=IDC_M_SQUARE_XYZ;
		if (strstr(code,"IDC_M_SQUARE_Z")) Conf->GridSelection=IDC_M_SQUARE_Z;
		if (strstr(code,"IDC_M_SQUARE_XZ")) Conf->GridSelection=IDC_M_SQUARE_XZ;
		if (strstr(code,"IDC_M_CUBE_TRIANGLE")) Conf->GridSelection=IDC_M_CUBE_TRIANGLE;
		if (strstr(code,"IDC_M_CUBE_WEB")) Conf->GridSelection=IDC_M_CUBE_WEB;

		fclose(filein);
	}
}

void SetMode()
{
	// Setup and activate the control devices
	
	switch(Conf->ControlMode)
	{
		case  IDC_CONTROL_JOYSTICK:	
				Mode->Control=Joystick;
		break;
		case IDC_CONTROL_KEYBOARD:
				Mode->Control=Keyboard;
		break;	
		default:
				Mode->Control=Keyboard;
		break;
	}
	
	switch(Conf->LineWidthCode)
	{
		case IDC_LINE1:
			Mode->LineWidth=1;
		break;
		case IDC_LINE2:
			Mode->LineWidth=2;
		break;
		case IDC_LINE3:
			Mode->LineWidth=3;
		break;
	}

	// translate OSD mode

	if (Conf->OSD_mode==IDC_OSD_YES)
	{
		Mode->OSD_visible=TRUE;
	}
	else
	{
		Mode->OSD_visible=FALSE;
	}
	
	switch(Conf->GridSelection)
	{
		case IDC_M_SQUARE_XYZ:
			Mode->GridType=MSquare_XYZ;
		break;
		case IDC_M_SQUARE_Z:
			Mode->GridType=MSquare_Z;
		break;
		case IDC_M_SQUARE_XZ:
			Mode->GridType=MSquare_XZ;
		break;
		case IDC_M_CUBE_TRIANGLE:
			Mode->GridType=MCube_Triangle;
		break;
		case IDC_M_CUBE_WEB:
			Mode->GridType=MCube_Web;
		break;	
	}
}
void SavePeaks()
{
	int i;

	FILE *fileout;
	
	fileout = fopen("newatoms.dat", "wt");				// File to write new peaks

	for (i=0;i<PeakData->Natoms;i++)
	{
		PeakData->a[i].fc=CalcCartToFract(PeakData->a[i].cc,
		MapDataS->PlaneTranslation,
		MapDataS->PlaneMatrix);
	}

	fprintf(fileout,"#EDIT\n");

	for (i=0;i<PeakData->Natoms;i++)
	{
		// CRYSTALS formated atom list
		fprintf(fileout,"ATOM Q  1   X=%10.6f  Y=%10.6f  Z=%10.6f \n",
		PeakData->a[i].fc.x,PeakData->a[i].fc.y,PeakData->a[i].fc.z);
	}

	fprintf(fileout,"END\n"); //CRYSTALS command

	fclose(fileout);

}

// CAVE specific coding coding

void init_shared_data(),gl_init_fn(),draw_fn(),idle_fn(),app_compute();

#ifndef CAVE_EMULATION
	#define SCALE_START 0.1
	#define SCALE_MAX 0.2
	#define SCALE_MIN 0.01
#else
	#define SCALE_START 20
	#define SCALE_MAX 100
	#define SCALE_MIN 10
#endif
	
main(int argc, char **argv)
{
	CAVEConfigure(&argc,argv,NULL);
	init_shared_data();
	CAVEInit();
	CAVEInitApplication(gl_init_fn,0);

#ifdef CAVE_EMULATION
	
	CAVE_PC_emu_idle(idle_fn);
	
	CAVEDisplay(draw_fn,0);
	
	return 1;
#else

	CAVEDisplay(draw_fn,0);

	idle_fn();
        
 	CAVEExit();
	
	return 1;
#endif
}

void idle_fn()
{
#ifdef CAVE_EMULATION
// pass throw GLUT emulation loop	
	if (! CAVEgetbutton(CAVE_ESCKEY))
	{
		app_compute();
	}
	else
	{
		SavePeaks();
		CAVEExit();
	}

#else 
// normal CAVE infinite loop
	while  (! CAVEgetbutton(CAVE_ESCKEY))
	{
		app_compute();
		sginap(1);
	}
	SavePeaks();
#endif	
}

// chose the nevigation response
int select_test( int joystick_test, int keyboard_test)
{
	if(Mode->Control==Joystick)
	{
		return joystick_test;
	}
	else
	{
		return keyboard_test;
	}	
}

void app_compute()
{
float jx,jy;
static int cbt1_down=FALSE;
static int cbt2_down=FALSE;
static int cbt3_down=FALSE;

// red the joystick only when it is enabled in configuration file
if(Mode->Control==Joystick)
{
	jx=CAVE_JOYSTICK_X;
	jy=CAVE_JOYSTICK_Y;
}

if(select_test(CAVEBUTTON1,CAVEgetbutton(CAVE_MKEY)))
 {
	if(!cbt1_down)
	{
		cbt1_down=TRUE;

// mode cycling
		if (*CurrentCaveMode==Translation)
		{
			*CurrentCaveMode=Rotation;
		}
		else if(*CurrentCaveMode==Rotation)
		{
			*CurrentCaveMode=Scaling;
		}
		else if(*CurrentCaveMode==Scaling)
		{
			*CurrentCaveMode=ELD_adjustment;
		}
		else if(*CurrentCaveMode==ELD_adjustment)
		{
			*CurrentCaveMode=ELD_on_off;
		}
// wee need cursor postioniong only in emulation mode
#ifdef CAVE_EMULATION
		else if(*CurrentCaveMode==ELD_on_off)
		{
			*CurrentCaveMode=Cursor3D;
		}
		else if(*CurrentCaveMode==Cursor3D)
		{
			*CurrentCaveMode=PeakManipulation;
		}
#else	// skipping cursor manipulation mode for true CAVE device
		else if(*CurrentCaveMode==ELD_on_off)
		{
			*CurrentCaveMode=PeakManipulation;
		}
#endif
		else
		{
			*CurrentCaveMode=Translation;
		}
	}
 }
 else
 {
	cbt1_down=FALSE;
 }

// we can adjust the joystick sensitivity in this place
#define JOYXLEFT jx < -0.99
#define JOYXRIGHT jx > 0.99
#define JOYYLEFT jy < -0.99
#define JOYYRIGHT jy > 0.99

 switch (*CurrentCaveMode)
 {
	case Translation:

		if(select_test((JOYXLEFT ),CAVEgetbutton(CAVE_QKEY)))
		{
			if (*x_cave> -100) *x_cave -= 1; 
		}
		if(select_test((JOYXRIGHT),CAVEgetbutton(CAVE_WKEY)))
		{
			if (*x_cave< 100) *x_cave += 1; 
		}
		if(select_test((JOYYLEFT ),CAVEgetbutton(CAVE_AKEY)))
		{
			if (*z_cave> -100) *z_cave -= 1; 
		}
		if(select_test((JOYYRIGHT),CAVEgetbutton(CAVE_SKEY)))
		{
			if (*z_cave< 100) *z_cave += 1; 
		}
		if(select_test(CAVEBUTTON2,CAVEgetbutton(CAVE_ZKEY)))
		{
			if (*y_cave> -100) *y_cave -= 1; 
		}
		if(select_test(CAVEBUTTON3,CAVEgetbutton(CAVE_XKEY)))
		{
			if (*y_cave< 100) *y_cave += 1; 
		}
	break;

	case Rotation:
		if(select_test((JOYYLEFT ),CAVEgetbutton(CAVE_AKEY)))
		{
			if (*be_cave> 0)
			{
				*be_cave -= 1; 
			}
			else
			{
				*be_cave = 359;
			}
		}
		if(select_test((JOYYRIGHT),CAVEgetbutton(CAVE_SKEY)))
		{
			if (*be_cave< 359)
			{
				*be_cave += 1; 
			}
			else
			{
				*be_cave = 0;
			}
		}
		if(select_test((JOYXLEFT ),CAVEgetbutton(CAVE_QKEY)))
		{
			if (*al_cave> 0)
			{
				*al_cave -= 1; 
			}
			else
			{
				*al_cave = 359;
			}
		}
		if(select_test((JOYXRIGHT),CAVEgetbutton(CAVE_WKEY)))
		{
			if (*al_cave< 359)
			{
				*al_cave += 1; 
			}
			else
			{
				*al_cave = 0;
			}
		}
		if(select_test(CAVEBUTTON2,CAVEgetbutton(CAVE_ZKEY)))
		{
			if (*ga_cave> 0)
			{
				*ga_cave -= 1; 
			}
			else
			{
				*ga_cave = 359;
			}
		}
		if(select_test(CAVEBUTTON3,CAVEgetbutton(CAVE_XKEY)))
		{
			if (*ga_cave< 359)
			{
				*ga_cave += 1; 
			}
			else
			{
				*ga_cave = 0;
			}
		}
	break;

	case Scaling:
		if(select_test(CAVEBUTTON2,CAVEgetbutton(CAVE_ZKEY)))
		{
			*model_scale=SCALE_START;
		}
		if(select_test((JOYYLEFT ),CAVEgetbutton(CAVE_AKEY)))
		{
			if((1.01*(*model_scale)) < SCALE_MAX)
			{
				*model_scale=1.01*(*model_scale); 
			}
			else
			{
				*model_scale=SCALE_MAX;
			}

		}
		if(select_test((JOYYRIGHT),CAVEgetbutton(CAVE_SKEY)))
		{
			if((0.99*(*model_scale)) > SCALE_MIN)
			{
				*model_scale=0.99*(*model_scale); 
			}
			else
			{
				*model_scale=SCALE_MIN;
			}
		}
	break;

	case ELD_adjustment:

		if(select_test((JOYYLEFT ),CAVEgetbutton(CAVE_AKEY)))
		{
			if (*zero<0.99)
			{
				*zero=*zero+0.001;
				*need_dl_recalc=TRUE;
			}
		}

		if(select_test((JOYYRIGHT),CAVEgetbutton(CAVE_SKEY)))
		{
			if (*zero>-0.99)
			{
				*zero=*zero-0.001;
				*need_dl_recalc=TRUE;
			}
		}
		if(select_test(CAVEBUTTON2,CAVEgetbutton(CAVE_ZKEY)))
		{
			*zero=(float)START;
			*need_dl_recalc=TRUE;
		}
	break;

	case ELD_on_off:
		if(select_test((JOYXLEFT ),CAVEgetbutton(CAVE_QKEY)))
		{
			layer_visibility[0]=FALSE;
			*need_dl_recalc=TRUE;
		}
		if(select_test((JOYXRIGHT),CAVEgetbutton(CAVE_WKEY)))
		{
			layer_visibility[0]=TRUE;
			*need_dl_recalc=TRUE; 
		}
		if(select_test((JOYYLEFT ),CAVEgetbutton(CAVE_AKEY)))
		{
			layer_visibility[1]=FALSE;
			*need_dl_recalc=TRUE;
		}
		if(select_test((JOYYRIGHT),CAVEgetbutton(CAVE_SKEY)))
		{
			layer_visibility[1]=TRUE;
			*need_dl_recalc=TRUE;
		}
		if(select_test(CAVEBUTTON2,CAVEgetbutton(CAVE_ZKEY)))
		{
			layer_visibility[2]=FALSE;
			*need_dl_recalc=TRUE;
		}
		if(select_test(CAVEBUTTON3,CAVEgetbutton(CAVE_XKEY)))
		{
			layer_visibility[2]=TRUE;
			*need_dl_recalc=TRUE;
		}

	break;

#define CROSS_SHIFT 0.1

	case Cursor3D:

// on true CAVE we will use the Wand position for this section instead of joystick
#ifdef CAVE_EMULATION

		if(select_test((JOYXLEFT ),CAVEgetbutton(CAVE_QKEY)))	
		{
			if (*x_cross> -100)
			{
				*x_cross -= CROSS_SHIFT;
			}
		}
		if(select_test((JOYXRIGHT),CAVEgetbutton(CAVE_WKEY)))	
		{
			if (*x_cross< 100)
			{
				*x_cross += CROSS_SHIFT;

			}
		}
		if(select_test((JOYYLEFT ),CAVEgetbutton(CAVE_AKEY)))	
		{
			if (*z_cross> -100)
			{
				*z_cross -= CROSS_SHIFT;
			}

		}
		if(select_test((JOYYRIGHT),CAVEgetbutton(CAVE_SKEY)))
		{
			if (*z_cross< 100)
			{
				*z_cross += CROSS_SHIFT;
			}
		}
		if(select_test(CAVEBUTTON2,CAVEgetbutton(CAVE_ZKEY)))
		{
			if (*y_cross> -100) *y_cross -= CROSS_SHIFT; 
		}
		if(select_test(CAVEBUTTON3,CAVEgetbutton(CAVE_XKEY)))
		{
			if (*y_cross< 100) *y_cross += CROSS_SHIFT; 
		}
#endif

	break;

	case PeakManipulation:
		if(select_test(CAVEBUTTON2,CAVEgetbutton(CAVE_ZKEY)))
		{
			if(!cbt2_down) // adding new peak
			{
				cbt2_down=TRUE;

				if(PeakData->Natoms<MAXATOMS)
				{
					PeakData->Natoms+=1;
					PeakData->a[PeakData->Natoms-1].cc.x=*x_cross;
					PeakData->a[PeakData->Natoms-1].cc.y=*y_cross;
					PeakData->a[PeakData->Natoms-1].cc.z=*z_cross;
				}		
			}
		}
		else
		{
			cbt2_down=FALSE;
		}
		if(select_test(CAVEBUTTON3,CAVEgetbutton(CAVE_XKEY)))
		{
			if(!cbt3_down) // deleting a peak
			{
				cbt3_down=TRUE;
				 if(PeakData->Natoms>0)
				{
					PeakData->Natoms-=1;
				}
			}
		}
		else
		{
			cbt3_down=FALSE;
		}
	break;
 }

}

void init_shared_data()
{
	x_cave = (float  *) CAVEMalloc (sizeof(float));
	y_cave = (float  *) CAVEMalloc (sizeof(float));
	z_cave = (float  *) CAVEMalloc (sizeof(float));
	x_cross = (float  *) CAVEMalloc (sizeof(float));
	y_cross = (float  *) CAVEMalloc (sizeof(float));
	z_cross = (float  *) CAVEMalloc (sizeof(float));
	al_cave = (float  *) CAVEMalloc (sizeof(float));
	be_cave = (float  *) CAVEMalloc (sizeof(float));
	ga_cave = (float  *) CAVEMalloc (sizeof(float));
    CurrentCaveMode =(CaveMode *) CAVEMalloc (sizeof(CaveMode));
	model_scale = (float  *) CAVEMalloc (sizeof(float));
	zero = (float  *) CAVEMalloc (sizeof(float));
	need_dl_recalc = (int  *) CAVEMalloc (sizeof(int));
	layer_visibility = (int *) CAVEMalloc(3* sizeof(int));

	Mode = (MceVrMode *) CAVEMalloc (sizeof(MceVrMode));
	Conf = (MceVrConfiguration *) CAVEMalloc (sizeof(MceVrConfiguration));
	
	PeakData = (MoleculeData *) CAVEMalloc (sizeof(MoleculeData));

	MapDataS = (MapDataShared *)CAVEMalloc(sizeof(MapDataShared));

	*x_cave=0;
	*y_cave=0;
	*z_cave=0;
	*x_cross=0;
	*y_cross=0;
	*z_cross=0;
	*al_cave=0;
	*be_cave=0;
	*ga_cave=0;
	*model_scale=SCALE_START;
	*CurrentCaveMode=Translation;
	*zero=(float)START;
	*need_dl_recalc=FALSE;
	layer_visibility[0]=TRUE;
	layer_visibility[1]=TRUE;
	layer_visibility[2]=TRUE;
	PeakData->Natoms=0;
}



void gl_init_fn()
{
float light_pos[]={-500,100,500,0};
	glLightfv(GL_LIGHT0,GL_POSITION,light_pos);
	glEnable(GL_LIGHT0);
	glEnable (GL_COLOR_MATERIAL);		// color surfaces of materials
	glEnable (GL_NORMALIZE); // set correct lighting after glScalef command
// generate display list

	// make precofiguration
	SetDefaultConf();
	SetMode();

    // load our world from disk
    SetupWorld();

	// generate display lists
	GenerateDisplayLists();
}

void printstring_glut(void *font,char *string)
{
 int len,i;
 len=(int)strlen(string);
 for(i=0;i<len;i++)
	 if(Mode->OSD_visible) // we can disable all text output in this way ....
	 {
		glutStrokeCharacter(font,string[i]);
	 }
}

void draw_fn()
{

char szBuffer[255];
void *font_type=GLUT_STROKE_ROMAN;

#ifndef CAVE_EMULATION
	float wand_pos[3];
	float cursor_pos[3];
	float wand_vect[3];
	float molecule_pos[3];


	CAVEGetPosition(CAVE_WAND,wand_pos);
	CAVEGetVector(CAVE_WAND_FRONT,wand_vect);
	
	//printf(" c1 %f c2 %f c3 %f",cursor_vect[0],cursor_vect[1],cursor_vect[2]);
	#define C_DIST 50
	cursor_pos[0]=wand_pos[0] + wand_vect[0]*C_DIST;
	cursor_pos[1]=wand_pos[1] + wand_vect[1]*C_DIST;
	cursor_pos[2]=wand_pos[2] + wand_vect[2]*C_DIST;
#endif

glDisable(GL_LIGHTING);
glEnable(GL_DEPTH_TEST);

	glClear(GL_DEPTH_BUFFER_BIT|GL_COLOR_BUFFER_BIT);
	glColor3f (1.0, 1.0, 1.0); // white
	
#ifndef CAVE_EMULATION
   
   glPushMatrix();
	glTranslatef(cursor_pos[0],cursor_pos[1],cursor_pos[2]);
	glScalef(5,5,5);
	NavigationCross();
   glPopMatrix();

   glPushMatrix();
	glBegin(GL_LINES);
		glVertex3f(wand_pos[0],wand_pos[1],wand_pos[2]);
		glVertex3f(cursor_pos[0],cursor_pos[1],cursor_pos[2]);
	glEnd();
   glPopMatrix();
#endif

#ifndef CAVE_EMULATION
	#define LETTER_SCALE 0.5
	#define L_SCALE 0.07*LETTER_SCALE
	#define L_SHIFT -10*LETTER_SCALE
	#define L_START 150
#else
   	#define LETTER_SCALE 1
	#define L_SCALE 0.07*LETTER_SCALE
	#define L_SHIFT -10*LETTER_SCALE
	#define L_START 140
#endif


// we need better visibility of the text ...
  glLineWidth(2)   ;
  glColor3f (1.0, 1.0, 1.0); // white

#ifdef CAVE_EMULATION
   glPushMatrix(); 	
	glTranslatef(-90,L_START,-120);
	glScalef(L_SCALE,L_SCALE, L_SCALE);
	printstring_glut(font_type,"MCE_CAVE (c) Michal Husak 2003");	
   glPopMatrix();
#endif

	switch(*CurrentCaveMode)
	{
	case Translation:
		glPushMatrix(); 
			glTranslatef(-90,L_START+L_SHIFT,-120);
			glScalef(L_SCALE,L_SCALE, L_SCALE);
			printstring_glut(font_type,"Translation mode");
		glPopMatrix();
		 glPushMatrix(); 
			glTranslatef(-90,L_START+2*L_SHIFT,-120);
			glScalef(L_SCALE,L_SCALE, L_SCALE);
			sprintf(szBuffer,"Position x %6.2f y %6.2f z %6.2f",*x_cave,*y_cave,*z_cave);
			printstring_glut(font_type,szBuffer);	
		glPopMatrix();
	break;
	case Rotation:
		glPushMatrix(); 
			glTranslatef(-90,L_START+L_SHIFT,-120);
			glScalef(L_SCALE,L_SCALE, L_SCALE);
			printstring_glut(font_type,"Rotation mode");
		glPopMatrix();
		glPushMatrix(); 
			glTranslatef(-90,L_START+2*L_SHIFT,-120);
			glScalef(L_SCALE,L_SCALE, L_SCALE);
			sprintf(szBuffer,"X axis rotation %6.2f",*al_cave);
			printstring_glut(font_type,szBuffer);
		glPopMatrix();
		glPushMatrix(); 
			glTranslatef(-90,L_START+3*L_SHIFT,-120);
			glScalef(L_SCALE,L_SCALE, L_SCALE);
			sprintf(szBuffer,"Y axis rotation %6.2f",*be_cave);
			printstring_glut(font_type,szBuffer);
		glPopMatrix();
		glPushMatrix(); 
			glTranslatef(-90,L_START+4*L_SHIFT,-120);
			glScalef(L_SCALE,L_SCALE, L_SCALE);
			sprintf(szBuffer,"Z axis rotation %6.2f",*ga_cave);
			printstring_glut(font_type,szBuffer);
		glPopMatrix();
	break;
	case Scaling:
		glPushMatrix(); 
			glTranslatef(-90,L_START+1*L_SHIFT,-120);
			glScalef(L_SCALE,L_SCALE, L_SCALE);
			printstring_glut(font_type,"Scaling mode");
		glPopMatrix();
		 glPushMatrix(); 
			glTranslatef(-90,L_START+2*L_SHIFT,-120);
			glScalef(L_SCALE,L_SCALE, L_SCALE);
			// normalizing the scale to be more intuitive
			float normalized_scale;
#ifndef CAVE_EMULATION
			normalized_scale=(SCALE_MAX-(*model_scale))/(SCALE_MAX-SCALE_MIN); // reverting value meaning
#else
			normalized_scale=((*model_scale)-SCALE_MIN)/(SCALE_MAX-SCALE_MIN);
#endif
			if (normalized_scale<0) normalized_scale =0; // to avoid rounding errorsmmmmmmmmmmmmmmmmm
			sprintf(szBuffer,"Model scale %6.2f",normalized_scale);
			printstring_glut(font_type,szBuffer);	
		glPopMatrix();
	break;
	case ELD_adjustment:
		glPushMatrix(); 
			glTranslatef(-90,L_START+1*L_SHIFT,-120);
			glScalef(L_SCALE,L_SCALE, L_SCALE);
			printstring_glut(font_type,"ELD_adjustment");
		glPopMatrix();
		 glPushMatrix(); 
			glTranslatef(-90,L_START+2*L_SHIFT,-120);
			glScalef(L_SCALE,L_SCALE, L_SCALE);
			sprintf(szBuffer,"Recalc coeffient %6.2f",*zero);
			printstring_glut(font_type,szBuffer);	
		glPopMatrix();
	break;
	case ELD_on_off:
		glPushMatrix(); 
			glTranslatef(-90,L_START+1*L_SHIFT,-120);
			glScalef(L_SCALE,L_SCALE, L_SCALE);
			printstring_glut(font_type,"ELD_on_off");
		glPopMatrix();
		 glPushMatrix(); 
			glTranslatef(-90,L_START+2*L_SHIFT,-120);
			glScalef(L_SCALE,L_SCALE, L_SCALE);
			sprintf(szBuffer,"Laeyr1 %d Layer2 %d Layer3 %d",layer_visibility[0],layer_visibility[1],layer_visibility[2]);
			printstring_glut(font_type,szBuffer);	
		glPopMatrix();
	break;
#ifdef CAVE_EMULATION
	// this sequence is nccesary ony for non-wand based cursor manipulation
	case Cursor3D:
		glPushMatrix(); 
			glTranslatef(-90,L_START+1*L_SHIFT,-120);
			glScalef(L_SCALE,L_SCALE, L_SCALE);
			printstring_glut(font_type,"Cursor position mode");
		glPopMatrix();
		 glPushMatrix(); 
			glTranslatef(-90,L_START+2*L_SHIFT,-120);
			glScalef(L_SCALE,L_SCALE, L_SCALE);
			sprintf(szBuffer,"Position x %6.2f y %6.2f z %6.2f",*x_cross,*y_cross,*z_cross);
			printstring_glut(font_type,szBuffer);	
		glPopMatrix();
	break;
#endif
	case PeakManipulation:
		glPushMatrix(); 
			glTranslatef(-90,L_START+1*L_SHIFT,-120);
			glScalef(L_SCALE,L_SCALE, L_SCALE);
			printstring_glut(font_type,"PeakManipulation mode");
		glPopMatrix();
		 glPushMatrix(); 
			glTranslatef(-90,L_START+2*L_SHIFT,-120);
			glScalef(L_SCALE,L_SCALE, L_SCALE);
			sprintf(szBuffer,"Current number of peaks %d",PeakData->Natoms);
			printstring_glut(font_type,szBuffer);	
		glPopMatrix();
	break;
	}

	// following operation coluld be alternatively shared for all process
   if(*need_dl_recalc)
   {
   		SetLayerLimits(*zero,(float)STEP,&layer_limits[0],&MapData);
		GenerateMesh();
		glDeleteLists(MCE_CONTOURS,1);
		GenerateDisplayLists();
		*need_dl_recalc=FALSE;
   }
 
#ifndef CAVE_EMULATION
   // for ture CAVE we use Nav operation to be able to recalulate
   // easy wand position into the molecule coordinates

  CAVENavLock();
  CAVENavLoadIdentity();

  CAVENavWorldTranslate(*x_cave,*y_cave,*z_cave);
  
  CAVENavWorldRot(*al_cave,'x');
  CAVENavWorldRot(*be_cave,'y');
  CAVENavWorldRot(*ga_cave,'z');
 
  CAVENavWorldScale(*model_scale,*model_scale,*model_scale);
  CAVENavTransform();
  CAVENavConvertCAVEToWorld(cursor_pos,molecule_pos);
  CAVENavUnlock();
  *x_cross = molecule_pos[0];
  *y_cross = molecule_pos[1];
  *z_cross = molecule_pos[2];
  
  glPushMatrix();
		DrawCAVEScene();
  glPopMatrix();
#else

glPushMatrix();	 
	glTranslatef(*x_cave,*y_cave,*z_cave);
	glRotatef(*al_cave,1,0,0);
	glRotatef(*be_cave,0,1,0);
	glRotatef(*ga_cave,0,0,1);
	glScalef(*model_scale,*model_scale,*model_scale);
			  glPushMatrix();
   				glTranslatef(*x_cross,*y_cross,*z_cross);
				NavigationCross();
				// cross is in the molecule coordinate system ....
			  glPopMatrix();
         DrawCAVEScene();
glPopMatrix();

#endif
}
