// CAVE PC emulator aplha 0.1 by Michal Husak 2001 husakm@vscht.cz
#include "cave_ogl_emul.h"
#include <stdio.h>      // Header file for standard file i/o.

// CAVE wall identification
typedef enum tagCaveWallIdentification
{CAVE_FRONT_WALL,CAVE_LEFT_WALL,CAVE_RIGHT_WALL,CAVE_FLOOR_WALL}
CAVE_WALL_ID;

// CAVE simlation view identification
typedef enum tagView_type
{Front,Left,Right,Floor,Outside}
View_type;

// CAVE visualisation mode
typedef enum tagVisual_mode
{mono,stereo,stereo_interlaced}
Visual_mode;

// CAVE parameters, this structure shuld be
// filed by data from cave.config file
typedef struct tagCaveDescription {
float Origin[3]; // distance to left wall, distance to floor, distance to front wall
float CAVEWidth; // Cave width & depth
float CAVEHeight; // Cave height
float xmin,xmax; // support variable to simplyfy the calulations
float ymin,ymax;
float zmin,zmax;
Visual_mode stereo_mode;
int CAVEEye_value;
} CaveDescription;


// Global variables
CaveDescription Cave;
View_type View = Outside;

GLint g_current_width, g_current_height; // get global acces to the window dimension
float g_aspect; // viewport aspect
int* transfer_argc; // command line parametere transfer
char** transfer_argv; // command line parametere transfer
unsigned int g_key_presed; // the key with is just down

// display functions registration pointers

void (*display_function)();
void (*initalize_function)();
void (*idle_function)();

int TestOpenGlError(const char * szMessage)
{
	GLenum errCode;
	const GLubyte *errString;
	char szBuffer[256];

	// OpenGl error examination
	if((errCode= glGetError())!=GL_NO_ERROR)
	{
		errString=gluErrorString(errCode);
		sprintf(szBuffer,"OpenGl Error: %s\n Error location: %s",errString,szMessage);	
#ifdef WIN32
		MessageBox(NULL,szBuffer,"OpenGL error",MB_ICONWARNING);
#else
			fprintf(stderr,szBuffer);
#endif
		return 0;
	}
	else
	{
		return 1;
	}
}

void CAVE_readstr(FILE *f, char *string)
{
    do {
	fgets(string, 255, f); // read the line
    } while ((string[0] == '/') || (string[0] == '\n'));
    return;
}

void InterpretCaverc()
{
	// overwrite CAVE defalut values by the values from .caverc
	FILE *filein;
	char oneline[255];
	char code[255];
	// testing
	int mode=1;
	

	filein = fopen(".caverc", "rt");				// File To Load World Data From
	if (filein == NULL)
	{
		fprintf(stderr,"no .caverc found, using default configuration\n");
		return;
	}

	while(! feof(filein))
	{
		CAVE_readstr(filein,oneline);
		sscanf(oneline, "DisplayMode %s\n", code);
		if (strstr(code,"stereo")) Cave.stereo_mode=stereo;
		if (strstr(code,"stereo_interlaced")) Cave.stereo_mode=stereo_interlaced;
	}
		
	fclose(filein);
}

void SetupCave()
{
	// set defalut cave values
	Cave.CAVEHeight=180;
	Cave.CAVEWidth=240;
	Cave.Origin[0]=120;
	Cave.Origin[1]=0;
	Cave.Origin[2]=120;

	Cave.stereo_mode=mono;
	Cave.CAVEEye_value=CAVE_LEFT_EYE;

	Cave.xmin=-Cave.CAVEWidth/2;
	Cave.xmax=Cave.CAVEWidth/2;

	Cave.ymin=0;
	Cave.ymax=Cave.CAVEHeight;

	Cave.zmin=-Cave.CAVEWidth/2;
	Cave.zmax=Cave.CAVEWidth/2;
}

void calculateFrustum(float eye[3], float cave[3], 
CAVE_WALL_ID wallID, float near_plane, float frustum[4])
{
  float d, l, r, b, t; // d - distnace from the procesed wall
  switch (wallID) {
  case CAVE_FRONT_WALL:
    d = 0.5f * cave[2] + eye[2]; // distance from front wall
    l = 0.5f * cave[0] + eye[0];
    r = 0.5f * cave[0] - eye[0];
    b = eye[1];
    t = cave[1] - eye[1];
    frustum[0] = -l * near_plane / d;
    frustum[1] =  r * near_plane / d;
    frustum[2] = -b * near_plane / d;
    frustum[3] =  t * near_plane / d;
    break;
  case CAVE_LEFT_WALL:
    d = 0.5f * cave[0] + eye[0];
    l = 0.5f * cave[2] + eye[2];
    r = 0.5f * cave[2] - eye[2];
    b = eye[1];
    t = cave[1] - eye[1];
    frustum[0] = -l * near_plane / d;
    frustum[1] =  r * near_plane / d;
    frustum[2] = -b * near_plane / d;
    frustum[3] =  t * near_plane / d;
    break;
  case CAVE_RIGHT_WALL:
    d = 0.5f * cave[0] - eye[0];
    l = 0.5f * cave[2] - eye[2];
    r = 0.5f * cave[2] + eye[2];
    b = eye[1];
    t = cave[1] - eye[1];
    frustum[0] = -l * near_plane / d;
    frustum[1] =  r * near_plane / d;
    frustum[2] = -b * near_plane / d;
    frustum[3] =  t * near_plane / d;
    break;
  case CAVE_FLOOR_WALL:
    d = eye[1];
    l = 0.5f * cave[0] + eye[0];
    r = 0.5f * cave[0] - eye[0];
    b = 0.5f * cave[2] + eye[2];
    t = 0.5f * cave[2] - eye[2]; 
    frustum[0] = -l * near_plane / d;
    frustum[1] =  r * near_plane / d;
    frustum[2] = -b * near_plane / d;
    frustum[3] =  t * near_plane / d;
    break;
  }
}

void DrawCaveWalls()
{
GLboolean bLightingState; // lighting state

	glGetBooleanv(GL_LIGHTING,&bLightingState);
	glDisable(GL_LIGHTING);

	glColor3f (1.0, 1.0, 1.0); // white lines
	// front wall
	glBegin(GL_LINE_STRIP);
		glVertex3f(Cave.xmin,Cave.ymin,Cave.zmin);
		glVertex3f(Cave.xmin,Cave.ymax,Cave.zmin);
		glVertex3f(Cave.xmax,Cave.ymax,Cave.zmin);
		glVertex3f(Cave.xmax,Cave.ymin,Cave.zmin);
	glEnd();

	// left wall
	glBegin(GL_LINE_STRIP);
		glVertex3f(Cave.xmin,Cave.ymin,Cave.zmin);
		glVertex3f(Cave.xmin,Cave.ymax,Cave.zmin);
		glVertex3f(Cave.xmin,Cave.ymax,Cave.zmax);
		glVertex3f(Cave.xmin,Cave.ymin,Cave.zmax);
	glEnd();

	// right wall
	glBegin(GL_LINE_STRIP);
		glVertex3f(Cave.xmax,Cave.ymin,Cave.zmin);
		glVertex3f(Cave.xmax,Cave.ymax,Cave.zmin);
		glVertex3f(Cave.xmax,Cave.ymax,Cave.zmax);
		glVertex3f(Cave.xmax,Cave.ymin,Cave.zmax);
	glEnd();

	// floor wall
	glColor3f (0.0, 1.0, 0.0); // green lines
	glBegin(GL_LINE_LOOP);
		glVertex3f(Cave.xmin,Cave.ymin,Cave.zmin);
		glVertex3f(Cave.xmax,Cave.ymin,Cave.zmin);
		glVertex3f(Cave.xmax,Cave.ymin,Cave.zmax);
		glVertex3f(Cave.xmin,Cave.ymin,Cave.zmax);
	glEnd();

	if(bLightingState) glEnable(GL_LIGHTING); // restore lighting state
}

void init(void) 
{
   SetupCave();
   InterpretCaverc();      
}

void Camera()
{
	const float add_view=1.1f; // aditonal view visible wall
	float eye[3]; // user eye position in cave (cm)
	#define EYE_SEPARATION 2
	float stereo_eye[3];
	float cave[4]; // cave diemnsions in cm
	CAVE_WALL_ID wallID; // identification of th user position in the cave
	float near_plane; //distance from the eye pos to nearest cliping plane cm
	float far_plane; //distance from the eye pos to far cliping plane cm
	float frustum[4]; //left, right, bottom und top of the glFrustrum command
	cave[0]= Cave.CAVEWidth ; //width of front desk ,x
	cave[1]= Cave.CAVEHeight ; //height of cave y
	cave[2]= Cave.CAVEWidth ; // width of  the side wallsz

	eye[0] =0;
	eye[1] =Cave.CAVEHeight/2;
	eye[2] =0;//Cave.CAVEHeight/2; // eye just in the cave center,  

	near_plane= 20 ; // 20 cm from eye will be the fisrt cuting plane
	far_plane= 600; // 600 cm far cliping plane

switch(Cave.stereo_mode)
{
case mono:
	switch(View)
	{
		case Outside:
			glMatrixMode (GL_PROJECTION);
			glLoadIdentity ();
			gluPerspective(45,g_aspect,10,1200);
			glMatrixMode (GL_MODELVIEW);
			glLoadIdentity ();             /* clear the matrix */
			gluLookAt(-400,200,600,0.0,Cave.CAVEHeight/2,0, 0.0, 1.0, 0.0);
		break;

		case Front:

			wallID=CAVE_FRONT_WALL; // front cave wall			
			calculateFrustum(eye,cave,wallID,near_plane,frustum);
			glMatrixMode (GL_PROJECTION);
			glLoadIdentity ();
			glFrustum(frustum[0]*add_view,
			    	  frustum[1]*add_view,
					  frustum[2]*add_view,
					  frustum[3]*add_view,
					  near_plane,far_plane);
			glMatrixMode (GL_MODELVIEW);
			glLoadIdentity ();             /* clear the matrix */
			gluLookAt(eye[0],eye[1],eye[2],0.0,Cave.CAVEHeight/2,Cave.zmin, 0.0, 1.0, 0.0);
		break;

		case Floor:

			wallID=CAVE_FLOOR_WALL; // floor cave wall
			calculateFrustum(eye,cave,wallID,near_plane,frustum);
			glMatrixMode (GL_PROJECTION);
			glLoadIdentity ();
			glFrustum(frustum[0]*add_view,
			    	  frustum[1]*add_view,
					  frustum[2]*add_view,
					  frustum[3]*add_view,
					  near_plane,far_plane);
			glMatrixMode (GL_MODELVIEW);
			glLoadIdentity ();             /* clear the matrix */
			gluLookAt(eye[0],eye[1],eye[2],0.0,0.0,0.0, 1.0, 0.0, 0.0);
		break;

		case Left:

			wallID=CAVE_LEFT_WALL; // left cave wall	
			calculateFrustum(eye,cave,wallID,near_plane,frustum);
			glMatrixMode (GL_PROJECTION);
			glLoadIdentity ();
			glFrustum(frustum[0]*add_view,
			    	  frustum[1]*add_view,
					  frustum[2]*add_view,
					  frustum[3]*add_view,
					  near_plane,far_plane);
			glMatrixMode (GL_MODELVIEW);
			glLoadIdentity ();             /* clear the matrix */
			gluLookAt(eye[0],eye[1],eye[2],Cave.xmin,Cave.CAVEHeight/2,0.0, 0.0, 1.0, 0.0);
		break;

		case Right:

			wallID=CAVE_RIGHT_WALL; // left cave wall			
			calculateFrustum(eye,cave,wallID,near_plane,frustum);
			glMatrixMode (GL_PROJECTION);
			glLoadIdentity ();
			glFrustum(frustum[0]*add_view,
			    	  frustum[1]*add_view,
					  frustum[2]*add_view,
					  frustum[3]*add_view,
					  near_plane,far_plane);
			glMatrixMode (GL_MODELVIEW);
			glLoadIdentity ();             /* clear the matrix */
			gluLookAt(eye[0],eye[1],eye[2],Cave.xmax,Cave.CAVEHeight/2,0.0, 0.0, 1.0, 0.0);
		break;
	}
break;
case stereo:

	switch(View)
	{
		case Outside:
			// Warning - the Outside stereoscopic projection is not 100 % correct !!!
			glMatrixMode (GL_PROJECTION);
			glLoadIdentity ();
			gluPerspective(45,g_aspect,10,1200);
			glMatrixMode (GL_MODELVIEW);
			glLoadIdentity ();             /* clear the matrix */
			if(Cave.CAVEEye_value==CAVE_LEFT_EYE)
			{
				gluLookAt(-410,200,600,0.0,Cave.CAVEHeight/2,0, 0.0, 1.0, 0.0);
			}
			else
			{
				gluLookAt(-390,200,600,0.0,Cave.CAVEHeight/2,0, 0.0, 1.0, 0.0);
			}

		break;

		case Front:
			if(Cave.CAVEEye_value==CAVE_LEFT_EYE)
			{
				stereo_eye[0]=eye[0]-EYE_SEPARATION;
				stereo_eye[1]=eye[1];
				stereo_eye[2]=eye[2];
			}
			else
			{
				stereo_eye[0]=eye[0]+EYE_SEPARATION;
				stereo_eye[1]=eye[1];
				stereo_eye[2]=eye[2];
			}
			wallID=CAVE_FRONT_WALL; // front cave wall			
			calculateFrustum(stereo_eye,cave,wallID,near_plane,frustum);
			glMatrixMode (GL_PROJECTION);
			glLoadIdentity ();
			glFrustum(frustum[0]*add_view,
			    	  frustum[1]*add_view,
					  frustum[2]*add_view,
					  frustum[3]*add_view,
					  near_plane,far_plane);
			glMatrixMode (GL_MODELVIEW);
			glLoadIdentity ();             /* clear the matrix */
			gluLookAt(stereo_eye[0],stereo_eye[1],stereo_eye[2],
			0.0,Cave.CAVEHeight/2,Cave.zmin, 0.0, 1.0, 0.0);
		break;

		case Floor:
			// I can not catch the stereo geometry correct :-(
			if(Cave.CAVEEye_value==CAVE_LEFT_EYE)
			{
				stereo_eye[0]=eye[0];
				stereo_eye[1]=eye[1];
				stereo_eye[2]=eye[2];
			}
			else
			{
				stereo_eye[0]=eye[0];
				stereo_eye[1]=eye[1];
				stereo_eye[2]=eye[2];
			}
			wallID=CAVE_FLOOR_WALL; // floor cave wall
			calculateFrustum(stereo_eye,cave,wallID,near_plane,frustum);
			glMatrixMode (GL_PROJECTION);
			glLoadIdentity ();
			glFrustum(frustum[0]*add_view,
			    	  frustum[1]*add_view,
					  frustum[2]*add_view,
					  frustum[3]*add_view,
					  near_plane,far_plane);
			glMatrixMode (GL_MODELVIEW);
			glLoadIdentity ();             /* clear the matrix */
			gluLookAt(stereo_eye[0],stereo_eye[1],stereo_eye[2],
			0.0,0.0,0.0, 1.0, 0.0, 0.0);

		break;

		case Left:
			if(Cave.CAVEEye_value==CAVE_LEFT_EYE)
			{
				stereo_eye[0]=eye[0];
				stereo_eye[1]=eye[1];
				stereo_eye[2]=eye[2]+EYE_SEPARATION;
			}
			else
			{
				stereo_eye[0]=eye[0];
				stereo_eye[1]=eye[1];
				stereo_eye[2]=eye[2]-EYE_SEPARATION;
			}

			wallID=CAVE_LEFT_WALL; // left cave wall	
			calculateFrustum(stereo_eye,cave,wallID,near_plane,frustum);
			glMatrixMode (GL_PROJECTION);
			glLoadIdentity ();
			glFrustum(frustum[0]*add_view,
			    	  frustum[1]*add_view,
					  frustum[2]*add_view,
					  frustum[3]*add_view,
					  near_plane,far_plane);
			glMatrixMode (GL_MODELVIEW);
			glLoadIdentity ();             /* clear the matrix */
			gluLookAt(stereo_eye[0],stereo_eye[1],stereo_eye[2],
			Cave.xmin,Cave.CAVEHeight/2,0.0, 0.0, 1.0, 0.0);
		break;

		case Right:
			if(Cave.CAVEEye_value==CAVE_LEFT_EYE)
			{
				stereo_eye[0]=eye[0];
				stereo_eye[1]=eye[1];
				stereo_eye[2]=eye[2]-EYE_SEPARATION;
			}
			else
			{
				stereo_eye[0]=eye[0];
				stereo_eye[1]=eye[1];
				stereo_eye[2]=eye[2]+EYE_SEPARATION;
			}

			wallID=CAVE_RIGHT_WALL; // left cave wall			
			calculateFrustum(stereo_eye,cave,wallID,near_plane,frustum);
			glMatrixMode (GL_PROJECTION);
			glLoadIdentity ();
			glFrustum(frustum[0]*add_view,
			    	  frustum[1]*add_view,
					  frustum[2]*add_view,
					  frustum[3]*add_view,
					  near_plane,far_plane);
			glMatrixMode (GL_MODELVIEW);
			glLoadIdentity ();             /* clear the matrix */
			gluLookAt(stereo_eye[0],stereo_eye[1],stereo_eye[2],Cave.xmax,Cave.CAVEHeight/2,0.0, 0.0, 1.0, 0.0);
			break;
	}
break;
}
}

void display(void)
{
switch (Cave.stereo_mode)
{
case mono:
	
	Cave.CAVEEye_value=CAVE_LEFT_EYE;

   Camera();

   glPushMatrix();
		display_function();
		TestOpenGlError("CAVE main mono mode");
   glPopMatrix();

   DrawCaveWalls();
		TestOpenGlError("CAVE wals mono mode");
   glFlush();

   glutSwapBuffers();
break;
case stereo:
	
	Cave.CAVEEye_value=CAVE_LEFT_EYE;

	glDrawBuffer(GL_BACK_LEFT);

   Camera();

   glPushMatrix();
		display_function();
		TestOpenGlError("CAVE main left eye");
   glPopMatrix();

   DrawCaveWalls();
		TestOpenGlError("CAVE wals left eye");

   Cave.CAVEEye_value=CAVE_RIGHT_EYE;

	glDrawBuffer(GL_BACK_RIGHT);

   Camera();

   glPushMatrix();
		display_function();
		TestOpenGlError("CAVE main right eye");
   glPopMatrix();

   DrawCaveWalls();
		TestOpenGlError("CAVE wals right eye");


   glFlush();

   glutSwapBuffers();
break;
}
}

void reshape (int w, int h)
{

	if (h==0)										// Prevent A Divide By Zero By
	{
		h=1;										// Making Height Equal One
	}

	g_current_width=w;
	g_current_height=h;

	g_aspect =(GLfloat)w/(GLfloat)h;	

   glViewport (0, 0, (GLsizei) w, (GLsizei) h); 

   Camera();
}

/* ARGSUSED1 */
void keyboard(unsigned char key, int x, int y)
{


  g_key_presed =key; // reamemeber the key for other process

   switch (key) {

	   // standard 0,1,2 Simulatro buttons used for CAVEBUTTONX emulation

	  case 'c': // look from outside
		  View=Outside;
		  glutPostRedisplay();
	  break;

	  case 'l': // left wall
		  View=Left;
		  glutPostRedisplay();
	  break;

	  case 'r': // right wall
		  View=Right;
		  glutPostRedisplay();
	  break;

	  case 'f': // front wall
		  View=Front;
		  glutPostRedisplay();
	  break;

	  case 'b': // button - floor
		  View=Floor;
		  glutPostRedisplay();
	  break;

// make this more exit options
/*
      case 27:
	      exit(0); 
      break;
*/
   }
}

void idle(void)
{
	idle_function();
	glutPostRedisplay();
}

int start_emulation()
{
   glutInit(transfer_argc, transfer_argv);
   switch  (Cave.stereo_mode)
   {
	case mono:
		glutInitDisplayMode (GLUT_SINGLE|GLUT_DOUBLE| GLUT_RGB);
	break;
	case stereo:
		glutInitDisplayMode (GLUT_SINGLE|GLUT_DOUBLE| GLUT_RGB | GLUT_STEREO);
	break;
	case stereo_interlaced:
		glutInitDisplayMode (GLUT_SINGLE|GLUT_DOUBLE| GLUT_RGB | GLUT_STENCIL);
	break;
   }
   glutInitWindowSize (500, 500); // optimized for 800x600 screen
   glutInitWindowPosition (150, 50);
   glutCreateWindow (transfer_argv[0]);
   initalize_function();
   glutDisplayFunc(display); 
   glutReshapeFunc(reshape);
   glutKeyboardFunc(keyboard);
   glutIdleFunc(idle);
   glutMainLoop();
   return 0;
}

void CAVEConfigure(int* argc,char** argv,void *test)
{
	transfer_argc=argc;
	transfer_argv=argv;
	init ();
}

void CAVEInit()
{
}
	
void CAVEInitApplication(void (*gl_init_fn)(),int i)
{
	initalize_function=gl_init_fn;
}

void CAVEDisplay(void (*draw_fn)(), int i)
{
	display_function=draw_fn;

	// folowing command shuld be solved by multithreading aproach 
	start_emulation(); // start the main GLUT code
}

	
void CAVEExit()
{
	exit(0);
}

void CAVE_PC_emu_idle(void (*idle_fn)())
{
	idle_function=idle_fn;
}

int CAVEgetbutton(unsigned int key)
{
if (g_key_presed!=0)
{	
	if (g_key_presed==key)
	{
		g_key_presed=0;
		return 1;
	}
}
	return 0;
}

int CAVEGetEye()
{
	return Cave.CAVEEye_value;
}

CheckJoysticButton(UINT button)
{

MMRESULT joyResult;
JOYINFO joy;

	joyResult= joyGetPos(JOYSTICKID1,&joy);
	
	if (joyResult == JOYERR_NOERROR  )
	{
		if (joy.wButtons == button) 
		{
			return 1;
		}
		else
		{
			return 0;
		}
	}
	else
	{
		MessageBox(NULL,"Joystick device failure, Recompile with keyboard support","Warning",MB_ICONERROR);

		return 0;
	}
}

float GiveJoysticValue(char direction)
{

MMRESULT joyResult;
JOYINFO joy;
float result=0;

	joyResult= joyGetPos(JOYSTICKID1,&joy);
	
	if (joyResult == JOYERR_NOERROR  )
	{
		switch (direction)
		{
		case 'X':
			result=(float)joy.wXpos/(float)65535;
		break;
		case 'Y':
			result=(float)joy.wYpos/(float)65535;
		break;
		case 'Z':
			result=(float)joy.wZpos/(float)65535;
		break;
		}
		return 2*(result-0.5); // recalulating to the <-1,1> range
	}
	else
	{
		MessageBox(NULL,"Joystick device failure, Recompile with keyboard support","Warning",MB_ICONERROR);
		return 0;
	}
}
